@@ -39,6 +39,7 @@ NSString *const VLCPlaybackServicePlaybackMetadataDidChange = @"VLCPlaybackServi
 NSString *const VLCPlaybackServicePlaybackDidFail = @"VLCPlaybackServicePlaybackDidFail";
 NSString *const VLCPlaybackServicePlaybackPositionUpdated = @"VLCPlaybackServicePlaybackPositionUpdated";
 NSString *const VLCPlaybackServicePlaybackModeUpdated = @"VLCPlaybackServicePlaybackModeUpdated";
+NSString *const VLCPlaybackServiceShuffleModeUpdated = @"VLCPlaybackServiceShuffleModeUpdated";
 NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackServicePlaybackDidMoveOnToNextItem";
 
 #if TARGET_OS_IOS
@@ -135,6 +136,8 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
         _playbackSessionManagementLock = [[NSLock alloc] init];
         _shuffleMode = NO;
+        _shuffledList = nil;
+        _shuffledOrder = [[NSMutableArray alloc] init];
 
         // Initialize a separate media player in order to play silence so that the application can
         // stay alive in background exclusively for Chromecast.
@@ -315,26 +318,23 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         APLog(@"%s: locking failed", __PRETTY_FUNCTION__);
         return;
     }
-    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
 
+    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
     BOOL equalizerEnabled = ![userDefaults boolForKey:kVLCSettingEqualizerProfileDisabled];
 
+    VLCAudioEqualizer *equalizer;
+
     if (equalizerEnabled) {
         NSArray *presets = [VLCAudioEqualizer presets];
         unsigned int profile = (unsigned int)[userDefaults integerForKey:kVLCSettingEqualizerProfile];
-        VLCAudioEqualizer *equalizer = [[VLCAudioEqualizer alloc] initWithPreset:presets[profile]];
-        equalizer.preAmplification = [userDefaults floatForKey:kVLCSettingDefaultPreampLevel];
-        _mediaPlayer.equalizer = equalizer;
+        equalizer = [[VLCAudioEqualizer alloc] initWithPreset:presets[profile]];
     } else {
         float preampValue = [userDefaults floatForKey:kVLCSettingDefaultPreampLevel];
-        if (preampValue != 0.) {
-            APLog(@"Enforcing presumbly disabled equalizer due to custom preamp value of %f2.0", preampValue);
-            VLCAudioEqualizer *equalizer = [[VLCAudioEqualizer alloc] init];
-            equalizer.preAmplification = preampValue;
-            _mediaPlayer.equalizer = equalizer;
-        }
+        equalizer = [[VLCAudioEqualizer alloc] init];
+        equalizer.preAmplification = preampValue;
     }
 
+    _mediaPlayer.equalizer = equalizer;
     [_mediaPlayer addObserver:self forKeyPath:@"time" options:0 context:nil];
 
 #if TARGET_OS_IOS
@@ -344,16 +344,13 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     /* we are playing a collection without a valid index,
      * so this is either 0 or totally random */
     if (_itemInMediaListToBePlayedFirst == -1) {
-        if (_shuffleMode) {
-            int count = (int)_mediaList.count;
-            if (count > 0) {
+        int count = (int)_mediaList.count;
+        if (_shuffleMode && count > 0) {
                 _currentIndex = arc4random_uniform(count - 1);
                 [self shuffleMediaList];
-                _itemInMediaListToBePlayedFirst = (int)[_shuffledOrder[0] integerValue];
-            }
-        } else {
-            _itemInMediaListToBePlayedFirst = 0;
         }
+
+        _itemInMediaListToBePlayedFirst = 0;
     }
 
     VLCMedia *media = [_mediaList mediaAtIndex:_itemInMediaListToBePlayedFirst];
@@ -431,6 +428,7 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         }
 
         _mediaPlayer = nil;
+        _shuffledList = nil;
         _listPlayer = nil;
     }
 
@@ -449,8 +447,10 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     [_playbackSessionManagementLock unlock];
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidStop object:self];
     if (_sessionWillRestart) {
-        _sessionWillRestart = NO;
-        [self startPlayback];
+        dispatch_async(dispatch_get_main_queue(), ^{
+            self->_sessionWillRestart = NO;
+            [self startPlayback];
+        });
     }
 }
 
@@ -461,7 +461,9 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
     if (media) {
         _mediaPlayer.currentAudioTrackIndex = (int) media.audioTrackIndex;
-        _mediaPlayer.currentVideoSubTitleIndex = (int) media.subtitleTrackIndex;
+
+        BOOL disableSubtitles = [[NSUserDefaults standardUserDefaults] boolForKey:kVLCSettingDisableSubtitles];
+        _mediaPlayer.currentVideoSubTitleIndex = disableSubtitles ? -1 : (int) media.subtitleTrackIndex;
     }
 }
 #endif
@@ -632,12 +634,20 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (NSInteger)numberOfAudioTracks
 {
+#if TARGET_OS_IOS
     return [_mediaPlayer numberOfAudioTracks] + 1;
+#else
+    return [_mediaPlayer numberOfAudioTracks];
+#endif
 }
 
 - (NSInteger)numberOfVideoSubtitlesIndexes
 {
-    return _mediaPlayer.videoSubTitlesIndexes.count + 2;
+#if TARGET_OS_IOS
+    return [_mediaPlayer numberOfSubtitlesTracks] + 2;
+#else
+    return [_mediaPlayer numberOfSubtitlesTracks] + 1;
+#endif
 }
 
 - (NSInteger)numberOfTitles
@@ -652,24 +662,27 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (NSString *)videoSubtitleNameAtIndex:(NSInteger)index
 {
-    NSInteger count = _mediaPlayer.videoSubTitlesNames.count;
-    if (index >= 0 && index < count) {
-        return _mediaPlayer.videoSubTitlesNames[index];
-    } else if (index == count) {
+    NSArray *videoSubTitlesNames = _mediaPlayer.videoSubTitlesNames;
+    NSInteger count = videoSubTitlesNames.count;
+
+    if (index == count) {
         return NSLocalizedString(@"SELECT_SUBTITLE_FROM_FILES", nil);
-    }
-    else if (index == count + 1) {
-        return NSLocalizedString(@"DOWNLOAD_SUBS_FROM_OSO", nil);
+    } else if (index < count) {
+        return videoSubTitlesNames[index];
     }
     return @"";
 }
 
 - (NSString *)audioTrackNameAtIndex:(NSInteger)index
 {
-    if (index >= 0 && index < _mediaPlayer.audioTrackNames.count)
-        return _mediaPlayer.audioTrackNames[index];
-    else if (index == _mediaPlayer.audioTrackNames.count)
+    NSArray *audioTrackNames = _mediaPlayer.audioTrackNames;
+    NSInteger count = audioTrackNames.count;
+
+    if (index == count) {
         return NSLocalizedString(@"SELECT_AUDIO_FROM_FILES", nil);
+    } else if (index < count) {
+        return audioTrackNames[index];
+    }
     return @"";
 }
 
@@ -690,16 +703,18 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (void)selectAudioTrackAtIndex:(NSInteger)index
 {
-    if (index >= 0 && index < _mediaPlayer.audioTrackIndexes.count) {
+    NSArray *audioTrackIndexes = _mediaPlayer.audioTrackIndexes;
+    if (index >= 0 && index < audioTrackIndexes.count) {
         //we can cast this cause we won't have more than 2 million audiotracks
-        _mediaPlayer.currentAudioTrackIndex = [_mediaPlayer.audioTrackIndexes[index] intValue];
+        _mediaPlayer.currentAudioTrackIndex = [audioTrackIndexes[index] intValue];
     }
 }
 
 - (void)selectVideoSubtitleAtIndex:(NSInteger)index
 {
-    if (index >= 0 && index < _mediaPlayer.videoSubTitlesIndexes.count) {
-        _mediaPlayer.currentVideoSubTitleIndex = [_mediaPlayer.videoSubTitlesIndexes[index] intValue];
+    NSArray *videoSubTitlesIndexes = _mediaPlayer.videoSubTitlesIndexes;
+    if (index >= 0 && index < videoSubTitlesIndexes.count) {
+        _mediaPlayer.currentVideoSubTitleIndex = [videoSubTitlesIndexes[index] intValue];
     }
 }
 
@@ -801,7 +816,7 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
             NSInteger nextIndex = [self nextMediaIndex:false];
 
             if (nextIndex > -1) {
-                if (_listPlayer.repeatMode != VLCRepeatCurrentItem && _shuffleMode) {
+                if (_listPlayer.repeatMode != VLCRepeatCurrentItem) {
                     [_listPlayer playItemAtNumber:@(nextIndex)];
                 }
                 [[NSNotificationCenter defaultCenter]
@@ -855,10 +870,12 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (void)playItemAtIndex:(NSUInteger)index
 {
-    VLCMedia *media = [_mediaList mediaAtIndex:index];
+    VLCMediaList *mediaList = _shuffleMode ? _shuffledList : _mediaList;
+    VLCMedia *media = [mediaList mediaAtIndex:index];
     [_listPlayer playItemAtNumber:[NSNumber numberWithUnsignedInteger:index]];
     _mediaPlayer.media = media;
-    _currentIndex = [_mediaList indexOfMedia:media];
+    _currentIndex = [mediaList indexOfMedia:media];
+
     if ([self.delegate respondsToSelector:@selector(prepareForMediaPlayback:)])
         [self.delegate prepareForMediaPlayback:self];
 }
@@ -870,14 +887,36 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     if (_shuffleMode) {
         [self shuffleMediaList];
         _currentIndex = 0;
+
+        @synchronized (_shuffledOrder) {
+            if ([_shuffledList count] == 0) {
+                NSMutableArray<VLCMedia *> *shuffledMedias = [[NSMutableArray alloc] init];
+                NSUInteger mediaListCount = _mediaList.count;
+                NSUInteger shuffledOrderCount = _shuffledOrder.count;
+                for (NSInteger i = _currentIndex; i < mediaListCount; i++) {
+                    if (i < shuffledOrderCount) {
+                        NSUInteger shuffleOrderIndex = [_shuffledOrder[i] unsignedIntegerValue];
+                        if (shuffleOrderIndex < mediaListCount) {
+                            [shuffledMedias addObject:[_mediaList mediaAtIndex:shuffleOrderIndex]];
+                        }
+                    }
+                }
+
+                _shuffledList = [[VLCMediaList alloc] initWithArray:shuffledMedias];
+                _listPlayer.mediaList = _shuffledList;
+            }
+        }
     } else {
-        _currentIndex = [_shuffledOrder[_currentIndex] integerValue];
+        _currentIndex = [_mediaList indexOfMedia:self.currentlyPlayingMedia];
+        _shuffledList = nil;
+        _listPlayer.mediaList = _mediaList;
     }
 
     if ([self.delegate respondsToSelector:@selector(playModeUpdated)]) {
         [self.delegate playModeUpdated];
     }
-    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
+
+    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServiceShuffleModeUpdated object:self];
 
     NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
     if ([[defaults valueForKey:kVLCPlayerShouldRememberState] boolValue]) {
@@ -886,21 +925,25 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 }
 
 - (void)shuffleMediaList {
-    NSInteger mediaListLength = _mediaList.count;
+    @synchronized (_shuffledOrder) {
+        NSInteger mediaListLength = _mediaList.count;
 
-    if (mediaListLength <= 1) {
-        return;
-    }
-    _shuffledOrder = [[NSMutableArray alloc]init];
-    for (int i = 0; i < mediaListLength; i++)
-    {
-        [_shuffledOrder addObject:[NSNumber numberWithInt:i]];
-    }
-    [_shuffledOrder exchangeObjectAtIndex:0 withObjectAtIndex:_currentIndex];
-    for (NSInteger i = 1; i < mediaListLength; i++) {
-        NSInteger nElements = mediaListLength - i;
-        NSInteger n = arc4random_uniform((uint32_t)nElements) + i;
-        [_shuffledOrder exchangeObjectAtIndex:i withObjectAtIndex:n];
+        if (mediaListLength <= 1) {
+            return;
+        }
+
+        [_shuffledOrder removeAllObjects];
+
+        for (int i = 0; i < mediaListLength; i++)
+        {
+            [_shuffledOrder addObject:[NSNumber numberWithInt:i]];
+        }
+        [_shuffledOrder exchangeObjectAtIndex:0 withObjectAtIndex:_currentIndex];
+        for (NSInteger i = 1; i < mediaListLength; i++) {
+            NSInteger nElements = mediaListLength - i;
+            NSInteger n = arc4random_uniform((uint32_t)nElements) + i;
+            [_shuffledOrder exchangeObjectAtIndex:i withObjectAtIndex:n];
+        }
     }
 }
 
@@ -910,18 +953,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
     NSInteger nextIndex = 0;
     if (!_currentIndex) {
-        _currentIndex = [_mediaList indexOfMedia:self.currentlyPlayingMedia];
+        VLCMediaList *currentMediaList = _shuffleMode ? _shuffledList : _mediaList;
+        _currentIndex = [currentMediaList indexOfMedia:self.currentlyPlayingMedia];
     }
 
-    // Change the repeat mode if next button is pressed
-    if (self.repeatMode == VLCRepeatCurrentItem && isButtonPressed) {
-        [self setRepeatMode:VLCRepeatAllItems];
-        if ([self.delegate respondsToSelector:@selector(playModeUpdated)]) {
-            [self.delegate playModeUpdated];
-        }
-
-        [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
-    } else if (self.repeatMode == VLCRepeatCurrentItem && !isButtonPressed) {
+    if (self.repeatMode == VLCRepeatCurrentItem && !isButtonPressed) {
         return _currentIndex;
     }
 
@@ -938,10 +974,6 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
     _currentIndex = nextIndex;
 
-    if (_shuffleMode && mediaListCount > 2 && _currentIndex >= 0 && _currentIndex < _shuffledOrder.count) {
-        return [_shuffledOrder[_currentIndex] integerValue];
-    }
-
     return _currentIndex;
 }
 
@@ -988,30 +1020,17 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
             [self savePlaybackState];
 #endif
             if (!_currentIndex) {
-                _currentIndex = [_mediaList indexOfMedia:self.currentlyPlayingMedia];
+                VLCMediaList *currentMediaList = _shuffleMode ? _shuffledList : _mediaList;
+                _currentIndex = [currentMediaList indexOfMedia:self.currentlyPlayingMedia];
             }
 
-            // Change the repeat mode if next button is pressed
-            if (self.repeatMode == VLCRepeatCurrentItem) {
-                [self setRepeatMode:VLCRepeatAllItems];
-                if ([self.delegate respondsToSelector:@selector(playModeUpdated)]) {
-                    [self.delegate playModeUpdated];
-                }
-                [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
-            }
-
-            if(_currentIndex > 0) {
+            if (_currentIndex > 0) {
                 _currentIndex -= 1;
-            } else{
-                if(_listPlayer.repeatMode == VLCRepeatAllItems) {
-                    _currentIndex = _mediaList.count - 1;
-                }
-            }
-            if(_shuffleMode){
-                [_listPlayer playItemAtNumber:@([_shuffledOrder[_currentIndex] integerValue])];
-            }else{
-                [_listPlayer playItemAtNumber:@(_currentIndex)];
+            } else if (_listPlayer.repeatMode == VLCRepeatAllItems) {
+                _currentIndex = _mediaList.count - 1;
             }
+
+            [_listPlayer playItemAtNumber:@(_currentIndex)];
         }
     } else {
         NSNumber *skipLength = [[NSUserDefaults standardUserDefaults] valueForKey:kVLCSettingPlaybackBackwardSkipLength];
@@ -1128,7 +1147,8 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     if (!enabled)
         _mediaPlayer.currentVideoTrackIndex = -1;
     else if (_mediaPlayer.currentVideoTrackIndex == -1) {
-        for (NSNumber *trackId in _mediaPlayer.videoTrackIndexes) {
+        NSArray *videoTrackIndexes = _mediaPlayer.videoTrackIndexes;
+        for (NSNumber *trackId in videoTrackIndexes) {
             if ([trackId intValue] != -1) {
                 _mediaPlayer.currentVideoTrackIndex = [trackId intValue];
                 break;
@@ -1163,7 +1183,7 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 }
 
 #pragma mark - 360 Support
-#if !TARGET_OS_TV
+
 - (BOOL)updateViewpoint:(CGFloat)yaw pitch:(CGFloat)pitch roll:(CGFloat)roll fov:(CGFloat)fov absolute:(BOOL)absolute
 {
     //adjusting the values
@@ -1234,7 +1254,6 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 #endif
     return -1;
 }
-#endif
 
 #pragma mark - equalizer
 
@@ -1556,6 +1575,13 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         [self.delegate prepareForMediaPlayback:self];
 }
 
+- (void)disableSubtitlesIfNeeded
+{
+    if ([[NSUserDefaults standardUserDefaults] boolForKey:kVLCSettingDisableSubtitles]) {
+        _mediaPlayer.currentVideoSubTitleIndex = -1;
+    }
+}
+
 - (void)scheduleSleepTimerWithInterval:(NSTimeInterval)timeInterval
 {
     if (_sleepTimer) {
@@ -1650,6 +1676,33 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     if (activePlaybackSession)
         [[VLCAppCoordinator sharedInstance].mediaLibraryService savePlaybackStateFrom:self];
 }
+
+- (BOOL)mediaListContains:(NSURL *)url
+{
+    for (int index = 0; index < _mediaList.count; index++) {
+        if ([[_mediaList mediaAtIndex:index].url isEqual:url]) {
+            return YES;
+        }
+    }
+
+    return NO;
+}
+
+- (void)removeMediaFromMediaListAtIndex:(NSUInteger)index
+{
+    BOOL deleteCurrentMedia = [_mediaList indexOfMedia:self.currentlyPlayingMedia] == index ? YES : NO;
+
+    [_mediaList removeMediaAtIndex:index];
+    [_delegate reloadPlayQueue];
+
+    if (deleteCurrentMedia && _mediaList.count == 1) {
+        _currentIndex = 0;
+        [_listPlayer playItemAtNumber:@(_currentIndex)];
+    } else if (deleteCurrentMedia) {
+        _currentIndex -= 1;
+        [self next];
+    }
+}
 #endif
 
 #pragma mark - Renderer
