@@ -13,6 +13,12 @@ import UIKit
 import AVFoundation
 import MediaPlayer
 
+enum PlayerSeekState {
+    case `default`
+    case forward
+    case backward
+}
+
 enum PlayerPanType {
     case none
     case brightness
@@ -20,6 +26,11 @@ enum PlayerPanType {
     case projection
 }
 
+enum PlayerSeekGestureType {
+    case tap
+    case swipe
+}
+
 class PlayerViewController: UIViewController {
     // MARK: - Slider Gesture Contol
 
@@ -65,9 +76,20 @@ class PlayerViewController: UIViewController {
     var playbackService: PlaybackService = PlaybackService.sharedInstance()
     var queueViewController: QueueViewController?
     var alertController: UIAlertController?
+
+    // MARK: Seek
     var seekForwardBy: Int = 0
     var seekBackwardBy: Int = 0
+    var numberOfGestureSeek: Int = 0
+    var totalSeekDuration: Int = 0
+    var seekForwardBySwipe: Int = 0
+    var seekBackwardBySwipe: Int = 0
+    var forwardBackwardEqual: Bool = true
+    var tapSwipeEqual: Bool = true
+    var numberOfTapSeek: Int = 0
+    var previousSeekState: PlayerSeekState = .default
 
+    // MARK: UI Elements
     lazy var statusLabel: VLCStatusLabel = {
         var statusLabel = VLCStatusLabel()
         statusLabel.isHidden = true
@@ -88,9 +110,24 @@ class PlayerViewController: UIViewController {
 
     lazy var mediaScrubProgressBar: MediaScrubProgressBar = {
         var mediaScrubProgressBar: MediaScrubProgressBar = MediaScrubProgressBar()
+        mediaScrubProgressBar.delegate = self
         return mediaScrubProgressBar
     }()
 
+    var abRepeatView: ABRepeatView?
+
+    lazy var aMark: ABRepeatMarkView = {
+        let aMark = ABRepeatMarkView(icon: UIImage(named: "abRepeatMarkerFill"))
+        aMark.translatesAutoresizingMaskIntoConstraints = false
+        return aMark
+    }()
+
+    lazy var bMark: ABRepeatMarkView = {
+        let bMark = ABRepeatMarkView(icon: UIImage(named: "abRepeatMarkerFill"))
+        bMark.translatesAutoresizingMaskIntoConstraints = false
+        return bMark
+    }()
+
     lazy var moreOptionsActionSheet: MediaMoreOptionsActionSheet = {
         var moreOptionsActionSheet = MediaMoreOptionsActionSheet()
         moreOptionsActionSheet.moreOptionsDelegate = self
@@ -243,7 +280,7 @@ class PlayerViewController: UIViewController {
 
     private var fov: CGFloat = 0
 
-    private var viewTranslation: CGPoint = CGPoint(x: 0, y: 0)
+    private var minimizationInitialCenter: CGPoint?
 
     private lazy var deviceMotion: DeviceMotion = {
         let deviceMotion = DeviceMotion()
@@ -251,6 +288,7 @@ class PlayerViewController: UIViewController {
         return deviceMotion
     }()
 
+    // MARK: Constants
     private let ZOOM_SENSITIVITY: CGFloat = 5
 
     private let screenPixelSize = CGSize(width: UIScreen.main.bounds.width,
@@ -298,6 +336,13 @@ class PlayerViewController: UIViewController {
         return minimizeGestureRecognizer
     }()
 
+    lazy var doubleTapGestureRecognizer: UITapGestureRecognizer = {
+        let doubleTapGestureRecognizer = UITapGestureRecognizer(target: self,
+                                                                action: #selector(handleDoubleTapGesture(_:)))
+        doubleTapGestureRecognizer.numberOfTapsRequired = 2
+        return doubleTapGestureRecognizer
+    }()
+
     // MARK: - Init
 
     @objc init(mediaLibraryService: MediaLibraryService, rendererDiscovererManager: VLCRendererDiscovererManager, playerController: PlayerController) {
@@ -329,11 +374,25 @@ class PlayerViewController: UIViewController {
             image = UIImage(named: "rendererFull")
         }
 
+        setupSeekDurations()
+
         mediaNavigationBar.updateDeviceButton(with: image, color: color)
 
+        if playerController.isRememberStateEnabled {
+            playbackService.isShuffleMode = playerController.isShuffleEnabled
+            playbackService.repeatMode = playerController.isRepeatEnabled
+        }
+
         view.transform = .identity
     }
 
+    override func viewWillLayoutSubviews() {
+        super.viewWillLayoutSubviews()
+
+        // Adjust the position of the AB Repeat marks if needed based on the device's orientation.
+        mediaScrubProgressBar.adjustABRepeatMarks(aMark: aMark, bMark: bMark)
+    }
+
     // MARK: - Public methods
 
     func showPopup(_ popupView: PopupView, with contentView: UIView, accessoryViewsDelegate: PopupViewAccessoryViewsDelegate? = nil) {
@@ -365,6 +424,7 @@ class PlayerViewController: UIViewController {
         pinchRecognizer.isEnabled = !disable
         leftSwipeRecognizer.isEnabled = !disable
         rightSwipeRecognizer.isEnabled = !disable
+        doubleTapGestureRecognizer.isEnabled = !disable
     }
     
     override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey: Any]?, context: UnsafeMutableRawPointer?) {
@@ -398,6 +458,14 @@ class PlayerViewController: UIViewController {
         // Minimize the player
     }
 
+    func updateShuffleState() {
+        playbackService.isShuffleMode = !playbackService.isShuffleMode
+    }
+
+    func updateRepeatMode() {
+        playbackService.toggleRepeatMode()
+    }
+
     // MARK: - Private methods
 
     private func jumpBackwards(_ interval: Int = 10) {
@@ -408,13 +476,41 @@ class PlayerViewController: UIViewController {
         playbackService.jumpForward(Int32(interval))
     }
 
+    private func executeSeekFromGesture(_ type: PlayerSeekGestureType) {
+        var hudString: String = ""
+
+        let currentSeek: Int
+        if numberOfGestureSeek > 0 {
+            currentSeek = type == .tap ? seekForwardBy : seekForwardBySwipe
+            totalSeekDuration = previousSeekState == .backward ? currentSeek : totalSeekDuration + currentSeek
+            previousSeekState = .forward
+        } else {
+            currentSeek = type == .tap ? seekBackwardBy : seekBackwardBySwipe
+            totalSeekDuration = previousSeekState == .forward ? -currentSeek : totalSeekDuration - currentSeek
+            previousSeekState = .backward
+        }
+
+        if totalSeekDuration > 0 {
+            hudString = "⇒ "
+            jumpForwards(currentSeek)
+        } else {
+            hudString = "⇐ "
+            jumpBackwards(currentSeek)
+        }
+
+        // Convert the time in seconds into milliseconds in order to the get the right VLCTime value.
+        let duration: VLCTime = VLCTime(number: NSNumber(value: abs(totalSeekDuration) * 1000))
+        hudString.append(duration.stringValue)
+        statusLabel.showStatusMessage(hudString)
+    }
+
     private func showIcon(button: UIButton) {
         UIView.animate(withDuration: 0.5, animations: {
             button.isHidden = false
         }, completion: nil)
     }
 
-    private func openOptionView(view: MediaPlayerActionSheetCellIdentifier) {
+    private func openOptionView(view: ActionSheetCellIdentifier) {
         present(moreOptionsActionSheet, animated: true, completion: {
             self.moreOptionsActionSheet.addView(view)
         })
@@ -460,6 +556,12 @@ class PlayerViewController: UIViewController {
         case optionsNavigationBar.sleepTimerButton:
             resetSleepTimer()
             return
+        case optionsNavigationBar.abRepeatButton:
+            resetABRepeat()
+            break
+        case optionsNavigationBar.abRepeatMarksButton:
+            resetABRepeatMarks(true)
+            break
         default:
             assertionFailure("VideoPlayerViewController: Unvalid button.")
         }
@@ -470,11 +572,18 @@ class PlayerViewController: UIViewController {
         let windowWidth: CGFloat = window.bounds.width
         let location: CGPoint = recognizer.location(in: window)
 
-        // Default or right side of the screen
-        var panType: PlayerPanType = .volume
+        // If minimization handler not ended yet, don't detect other gestures to don't block it.
+        guard minimizationInitialCenter == nil else { return .none }
 
-        if location.x < windowWidth / 2 {
-            panType = .brightness
+        var panType: PlayerPanType = .none
+        if location.x < 3 * windowWidth / 3 && playerController.isVolumeGestureEnabled {
+            panType = .volume
+        }
+        if location.x < 2 * windowWidth / 3 {
+            panType = .none
+        }
+        if location.x < 1 * windowWidth / 3 && playerController.isBrightnessGestureEnabled {
+             panType = .brightness
         }
 
         if playbackService.currentMediaIs360Video {
@@ -524,6 +633,30 @@ class PlayerViewController: UIViewController {
         AVAudioSession.sharedInstance().addObserver(self, forKeyPath: "outputVolume", options: NSKeyValueObservingOptions.new, context: nil)
     }
 
+    private func setupSeekDurations() {
+        let defaults = UserDefaults.standard
+
+        tapSwipeEqual = defaults.bool(forKey: kVLCSettingPlaybackTapSwipeEqual)
+        forwardBackwardEqual = defaults.bool(forKey: kVLCSettingPlaybackForwardBackwardEqual)
+        seekForwardBy = defaults.integer(forKey: kVLCSettingPlaybackForwardSkipLength)
+        seekBackwardBy = forwardBackwardEqual ? seekForwardBy : defaults.integer(forKey: kVLCSettingPlaybackBackwardSkipLength)
+        seekForwardBySwipe = tapSwipeEqual ? seekForwardBy : defaults.integer(forKey: kVLCSettingPlaybackForwardSkipLengthSwipe)
+
+        if tapSwipeEqual, forwardBackwardEqual {
+            // if tap = swipe, and backward = forward, then backward swipe = forward tap
+            seekBackwardBySwipe = seekForwardBy
+        } else if tapSwipeEqual, !forwardBackwardEqual {
+            // if tap = swipe, and backward != forward, then backward swipe = backward tap
+            seekBackwardBySwipe = seekBackwardBy
+        } else if !tapSwipeEqual, forwardBackwardEqual {
+            // if tap != swipe, and backward = forward, then backward swipe = forward swipe
+            seekBackwardBySwipe = seekForwardBySwipe
+        } else {
+            // otherwise backward swipe = backward swipe
+            seekBackwardBySwipe = defaults.integer(forKey: kVLCSettingPlaybackBackwardSkipLengthSwipe)
+        }
+    }
+
     // MARK: - Gesture handlers
 
     @objc func handlePlayPauseGesture() {
@@ -545,6 +678,9 @@ class PlayerViewController: UIViewController {
         let currentPos = recognizer.location(in: view)
 
         let panType = detectPanType(recognizer)
+        if panType == .none {
+            return handleMinimizeGesture(recognizer)
+        }
 
         guard panType == .projection
                 || (panType == .volume && playerController.isVolumeGestureEnabled)
@@ -679,22 +815,13 @@ class PlayerViewController: UIViewController {
 
         switch recognizer.direction {
         case .right:
-            let timeRemaining = -Int(Double(playbackService.remainingTime().intValue) * 0.001)
-
-            if seekForwardBy < timeRemaining {
-                if seekForwardBy < 1 {
-                    seekForwardBy = 1
-                }
-
-                jumpForwards(seekForwardBy)
-                hudString = String(format: "⇒ %is", seekForwardBy)
-            } else {
-                jumpForwards(timeRemaining - 5)
-                hudString = String(format: "⇒ %is", (timeRemaining - 5))
-            }
+            numberOfGestureSeek = previousSeekState == .backward ? 1 : numberOfGestureSeek + 1
+            executeSeekFromGesture(.swipe)
+            return
         case .left:
-            jumpBackwards(seekBackwardBy)
-            hudString = String(format: "⇐ %is", seekBackwardBy)
+            numberOfGestureSeek = previousSeekState == .forward ? -1 : numberOfGestureSeek - 1
+            executeSeekFromGesture(.swipe)
+            return
         case .up:
             playbackService.previous()
             hudString = NSLocalizedString("BWD_BUTTON", comment: "")
@@ -710,44 +837,44 @@ class PlayerViewController: UIViewController {
         }
     }
 
-    @objc private func handleMinimizeGesture(_ sender: UIPanGestureRecognizer) {
-        switch sender.state {
+    @objc private func handleMinimizeGesture(_ gesture: UIPanGestureRecognizer) {
+        let translation = gesture.translation(in: view)
+
+        switch gesture.state {
+        case .began:
+            // Save the initial center position of the view
+            minimizationInitialCenter = view.center
+
         case .changed:
-            viewTranslation = sender.translation(in: view)
-            if viewTranslation.y < 0 {
-                return
+            // Move the view with the pan gesture
+            if let initialCenter = minimizationInitialCenter, translation.y > 0 { // Only allow downward sliding
+                view.center = CGPoint(x: view.center.x, y: initialCenter.y + translation.y)
             }
 
-            UIView.animate(withDuration: 0.5,
-                           delay: 0,
-                           usingSpringWithDamping: 0.7,
-                           initialSpringVelocity: 1,
-                           options: .curveEaseOut,
-                           animations: {
-                self.view.transform = CGAffineTransform(translationX: 0, y: self.viewTranslation.y)
-            })
-
-            break
-        case .ended:
-            let halfHeight = view.frame.height / 2
-            if viewTranslation.y < halfHeight {
-                UIView.animate(withDuration: 0.5,
-                               delay: 0,
-                               usingSpringWithDamping: 0.7,
-                               initialSpringVelocity: 1,
-                               options: .curveEaseOut,
-                               animations: {
-                    self.view.transform = .identity
-                })
-            } else {
-                minimizePlayer()
+        case .ended, .cancelled, .failed:
+            // Determine if the view should be dismissed or return to its original position
+            if let initialCenter = minimizationInitialCenter {
+                if translation.y > view.bounds.height * 0.1 { // Adjust the threshold as needed
+                    minimizePlayer()
+                } else {
+                    // Animate the view returning to its original position
+                    UIView.animate(withDuration: 0.3) {
+                        self.view.center = initialCenter
+                    }
+                }
             }
 
-            break
+            minimizationInitialCenter = nil
         default:
             break
         }
     }
+
+    @objc func handleDoubleTapGesture(_ sender: UITapGestureRecognizer) {
+        // CHECK THE TAP LOCATION
+
+        executeSeekFromGesture(.tap)
+    }
 }
 
 // MARK: - VLCPlaybackServiceDelegate
@@ -792,6 +919,12 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
         case .equalizer:
             showIcon(button: optionsNavigationBar.equalizerButton)
             break
+        case .abRepeat:
+            showIcon(button: optionsNavigationBar.abRepeatButton)
+            break
+        case .abRepeatMarks:
+            showIcon(button: optionsNavigationBar.abRepeatMarksButton)
+            break
         default:
             assertionFailure("AudioPlayerViewController: Invalid option.")
         }
@@ -808,6 +941,12 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
         case .equalizer:
             hideIcon(button: optionsNavigationBar.equalizerButton)
             break
+        case .abRepeat:
+            hideIcon(button: optionsNavigationBar.abRepeatButton)
+            break
+        case .abRepeatMarks:
+            hideIcon(button: optionsNavigationBar.abRepeatMarksButton)
+            break
         default:
             assertionFailure("AudioPlayerViewController: Invalid option.")
         }
@@ -855,7 +994,7 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
     }
 
     func mediaMoreOptionsActionSheetDisplayAlert(title: String, message: String, action: BookmarkActionIdentifier, index: Int, isEditing: Bool) {
-        moreOptionsActionSheet.dismiss(animated: true, completion: {
+        let completion = {
             let alertController = UIAlertController(title: title, message: message, preferredStyle: .alert)
 
             var actionTitle = NSLocalizedString("BUTTON_CANCEL", comment: "")
@@ -905,7 +1044,15 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
 
             self.present(alertController, animated: true, completion: nil)
             self.alertController = alertController
-        })
+        }
+
+        // iOS 12.0 and below versions do not execute the completion if the dismiss call is not performed,
+        // here the check is necessary in order to enable the edit actions for these iOS versions.
+        if addBookmarksView == nil {
+            moreOptionsActionSheet.dismiss(animated: true, completion: completion)
+        } else {
+            completion()
+        }
     }
 
     func mediaMoreOptionsActionSheetDisplayAddBookmarksView(_ bookmarksView: AddBookmarksView) {
@@ -931,24 +1078,40 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
     }
 
     func mediaMoreOptionsActionSheetDidToggleShuffle(_ mediaMoreOptionsActionSheet: MediaMoreOptionsActionSheet) {
-        playbackService.isShuffleMode = !playbackService.isShuffleMode
-
-        if playerController.isRememberStateEnabled {
-            UserDefaults.standard.setValue(playbackService.isShuffleMode, forKey: kVLCPlayerIsShuffleEnabled)
-        }
+        updateShuffleState()
 
         mediaMoreOptionsActionSheet.collectionView.reloadData()
     }
 
     func mediaMoreOptionsActionSheetDidTapRepeat(_ mediaMoreOptionsActionSheet: MediaMoreOptionsActionSheet) {
-        playbackService.toggleRepeatMode()
-
-        if playerController.isRememberStateEnabled {
-            UserDefaults.standard.setValue(playbackService.repeatMode.rawValue, forKey: kVLCPlayerIsRepeatEnabled)
-        }
+        updateRepeatMode()
 
         mediaMoreOptionsActionSheet.collectionView.reloadData()
     }
+
+    func mediaMoreOptionsActionSheetPresentABRepeatView(with abView: ABRepeatView) {
+        abView.translatesAutoresizingMaskIntoConstraints = false
+        abRepeatView = abView
+
+        guard let abRepeatView = abRepeatView else {
+            return
+        }
+
+        abRepeatView.aMarkView.isHidden = false
+        mediaScrubProgressBar.shouldHideScrubLabels = true
+    }
+
+    func mediaMoreOptionsActionSheetDidSelectAMark() {
+        mediaScrubProgressBar.setMark(aMark)
+        aMark.isEnabled = true
+    }
+
+    func mediaMoreOptionsActionSheetDidSelectBMark() {
+        mediaScrubProgressBar.setMark(bMark)
+        bMark.isEnabled = true
+        mediaScrubProgressBar.shouldHideScrubLabels = false
+        abRepeatView?.removeFromSuperview()
+    }
 }
 
 // MARK: - OptionsNavigationBarDelegate
@@ -973,6 +1136,46 @@ extension PlayerViewController: OptionsNavigationBarDelegate {
     func optionsNavigationBarGetRemainingTime() -> String {
         return moreOptionsActionSheet.getRemainingTime()
     }
+
+    func resetABRepeat() {
+        hideIcon(button: optionsNavigationBar.abRepeatButton)
+        if let abRepeatView = abRepeatView {
+            abRepeatView.removeFromSuperview()
+        }
+        resetABRepeatMarks()
+    }
+
+    private func resetABRepeatMarks(_ shouldDisplayView: Bool = false) {
+        hideIcon(button: optionsNavigationBar.abRepeatMarksButton)
+        aMark.removeFromSuperview()
+        aMark.isEnabled = false
+
+        bMark.removeFromSuperview()
+        bMark.isEnabled = false
+
+        guard let abRepeatView = abRepeatView,
+              shouldDisplayView else {
+            return
+        }
+
+        mediaMoreOptionsActionSheetPresentABRepeatView(with: abRepeatView)
+    }
+}
+
+// MARK: - MediaScrubProgressBarDelegate
+
+extension PlayerViewController: MediaScrubProgressBarDelegate {
+    func mediaScrubProgressBarSetPlaybackPosition(to value: Float) {
+        playbackService.playbackPosition = value
+    }
+
+    func mediaScrubProgressBarGetAMark() -> ABRepeatMarkView {
+        return aMark
+    }
+
+    func mediaScrubProgressBarGetBMark() -> ABRepeatMarkView {
+        return bMark
+    }
 }
 
 // MARK: - PopupViewDelegate
