@@ -43,6 +43,7 @@ class QueueViewController: UIViewController {
     @IBOutlet weak var closeButton: UIButton!
 
     private var scrolledCellIndex: IndexPath = IndexPath()
+    private var grabbedCellIndex: IndexPath?
 
     private let cellHeight: CGFloat = 56
 
@@ -59,9 +60,10 @@ class QueueViewController: UIViewController {
             PlaybackService.sharedInstance()
         }
     }
+
     private var mediaList: VLCMediaList {
         get {
-            PlaybackService.sharedInstance().mediaList
+            playbackService.isShuffleMode ? PlaybackService.sharedInstance().shuffledList : PlaybackService.sharedInstance().mediaList
         }
     }
 
@@ -101,14 +103,20 @@ class QueueViewController: UIViewController {
     override func viewDidLoad() {
         super.viewDidLoad()
         initViews()
-        NotificationCenter.default.addObserver(self,
-                                               selector: #selector(themeDidChange),
-                                               name: .VLCThemeDidChangeNotification,
-                                               object: nil)
-        NotificationCenter.default.addObserver(self,
-                                               selector: #selector(deviceOrientationDidChange),
-                                               name: UIDevice.orientationDidChangeNotification,
-                                               object: nil)
+
+        let defaultCenter: NotificationCenter = NotificationCenter.default
+        defaultCenter.addObserver(self,
+                                  selector: #selector(themeDidChange),
+                                  name: .VLCThemeDidChangeNotification,
+                                  object: nil)
+        defaultCenter.addObserver(self,
+                                  selector: #selector(deviceOrientationDidChange),
+                                  name: UIDevice.orientationDidChangeNotification,
+                                  object: nil)
+        defaultCenter.addObserver(self,
+                                  selector: #selector(reload),
+                                  name: Notification.Name(VLCPlaybackServiceShuffleModeUpdated),
+                                  object: nil)
     }
 
     override func viewWillDisappear(_ animated: Bool) {
@@ -161,7 +169,7 @@ class QueueViewController: UIViewController {
                     bottomConstraint = view.bottomAnchor.constraint(equalTo: parent.view.bottomAnchor,
                                                                     constant: self.view.frame.height)
                 } else if let parent = parent as? AudioPlayerViewController {
-                    heightConstraint = view.heightAnchor.constraint(equalToConstant: parent.audioPlayerView.playqueueView.bounds.height)
+                    heightConstraint = nil
                     grabberView.isHidden = true
                     closeButton.isHidden = true
                     topConstraint = view.topAnchor.constraint(equalTo: parent.audioPlayerView.playqueueView.topAnchor)
@@ -220,6 +228,27 @@ class QueueViewController: UIViewController {
 
         topConstraint?.constant = topConstraintConstant
         reload()
+
+        guard let currentMedia = playbackService.currentlyPlayingMedia else {
+            return
+        }
+
+        let currentIndex = mediaList.index(of: currentMedia)
+
+        guard currentIndex != NSNotFound else {
+            return
+        }
+
+        guard currentIndex < mediaList.count else {
+            return
+        }
+
+        let currentIndexPath = IndexPath(row: Int(currentIndex), section: 0)
+
+        // Dispatch the scrolling operation to the main queue to ensure UI updates are processed first
+        DispatchQueue.main.async {
+            self.queueCollectionView.scrollToItem(at: currentIndexPath, at: .centeredVertically, animated: true)
+        }
     }
 
     @objc init(medialibraryService: MediaLibraryService) {
@@ -418,9 +447,6 @@ private extension QueueViewController {
         if isSelected {
             textColor = PresentationTheme.current.colors.orangeUI
             tintColor = PresentationTheme.current.colors.orangeUI
-            cell.disableScrollView()
-        } else {
-            cell.enableScrollView()
         }
 
         cell.tintColor = tintColor
@@ -443,12 +469,31 @@ private extension QueueViewController {
                     break
             }
             queueCollectionView.beginInteractiveMovementForItem(at: selectedIndexPath)
+            grabbedCellIndex = selectedIndexPath
         case .changed:
             var location = gesture.location(in: gesture.view)
             location.x = queueCollectionView.frame.width / 2
             queueCollectionView.updateInteractiveMovementTargetPosition(location)
+            if let selectedIndexPath = queueCollectionView.indexPathForItem(at: gesture.location(in: queueCollectionView)) {
+                grabbedCellIndex = selectedIndexPath
+            }
         case .ended:
             queueCollectionView.endInteractiveMovement()
+            var indexPath: IndexPath? = nil
+
+            if let selectedIndexPath = queueCollectionView.indexPathForItem(at: gesture.location(in: queueCollectionView)) {
+                indexPath = selectedIndexPath
+            } else if let grabbedCellIndex = grabbedCellIndex {
+                indexPath = grabbedCellIndex
+            }
+
+            guard let index = indexPath,
+                  let cell = queueCollectionView.cellForItem(at: index) as? MediaCollectionViewCell else {
+                break
+            }
+
+            cell.animateCurrentlyPlayingState()
+            break
         default:
             queueCollectionView.cancelInteractiveMovement()
         }
@@ -491,17 +536,16 @@ extension QueueViewController: UICollectionViewDelegate, MediaCollectionViewCell
             assertionFailure("QueueViewController: didSelectItemAt: IndexPath out of range.")
             return
         }
-        let media = mediaList.media(at: UInt(indexPath.row))
-        let isPlaying = playbackService.currentlyPlayingMedia == media
-        if !isPlaying {
-            playbackService.playItem(at: UInt(indexPath.row))
-            guard let cell = collectionView.cellForItem(at: indexPath) as? MediaCollectionViewCell else {
-                assertionFailure("QueueViewController: didSelectItemAt: Cell not a MediaCollectionViewCell")
-                return
-            }
-            updateCollectionViewCellApparence(cell, isSelected: true)
-            reload()
+
+        playbackService.playItem(at: UInt(indexPath.row))
+
+        guard let cell = collectionView.cellForItem(at: indexPath) as? MediaCollectionViewCell else {
+            assertionFailure("QueueViewController: didSelectItemAt: Cell not a MediaCollectionViewCell")
+            return
         }
+
+        updateCollectionViewCellApparence(cell, isSelected: true)
+        reload()
     }
 
     func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
